@@ -56,16 +56,12 @@ try:
 except ModuleNotFoundError:
     _DO_FEED = False
 
-try:
-    from PIL import Image
-    _HAS_PIL = True
-except ModuleNotFoundError:
-    _HAS_PIL = False
 _HAS_TIMG = shutil.which('timg')
 _HAS_CHAFA = shutil.which('chafa')
 _NEW_CHAFA = False
 _NEW_TIMG = False
 _RENDER_IMAGE = False
+_HAS_PIL = False
 
 # All this code to know if we render image inline or not
 if _HAS_CHAFA:
@@ -91,8 +87,13 @@ if _HAS_TIMG :
     if output and output[5:10] > "1.3.2":
         _NEW_TIMG = True
         _RENDER_IMAGE = True
-elif _HAS_CHAFA and _HAS_PIL:
-    _RENDER_IMAGE = True
+elif _HAS_CHAFA and not _NEW_CHAFA:
+    try:
+        from PIL import Image
+        _HAS_PIL = True
+        _RENDER_IMAGE = True
+    except ModuleNotFoundError:
+        _HAS_PIL = False
 if not _RENDER_IMAGE:
     print("To render images inline, you need either chafa or timg.")
     if not _NEW_CHAFA and not _NEW_TIMG:
@@ -117,6 +118,8 @@ def inline_image(img_file,width):
     if _HAS_CHAFA:
         if _HAS_PIL and not _NEW_CHAFA:
             # this code is a hack to remove frames from animated gif
+            print("WARNING: support for chafa < 1.10 will soon be removed")
+            print("If you can’t upgrade chafa or timg, please contact offpunk developers")
             img_obj = Image.open(img_file)
             if hasattr(img_obj,"n_frames") and img_obj.n_frames > 1:
                 # we remove all frames but the first one
@@ -141,9 +144,9 @@ def terminal_image(img_file):
     # it is also centered
     cmds = []
     if _NEW_CHAFA:
-        cmds.append("chafa -C on -d 0 --bg white -t 1 -w 1")
+        cmds.append("chafa -C on -d 0 --bg white -w 1")
     elif _HAS_CHAFA:
-        cmds.append("chafa -d 0 --bg white -t 1 -w 1")
+        cmds.append("chafa -d 0 --bg white -w 1")
     if _NEW_TIMG:
         cmds.append("timg --loops=1 -C")
     image_success = False
@@ -727,6 +730,13 @@ class GemtextRenderer(AbstractRenderer):
         links += hidden_links
         return r.get_final(), links
 
+class EmptyRenderer(GemtextRenderer):
+    def get_mime(self):
+        return "text/empty"
+    def prepare(self,body,mode=None):
+        text= "(empty file)"
+        return [[text, "GemtextRenderer"]]
+
 class GopherRenderer(AbstractRenderer):
     def get_mime(self):
         return "text/gopher"
@@ -776,22 +786,25 @@ class GopherRenderer(AbstractRenderer):
                     parts = parts[:-1]
                 if len(parts) == 4:
                     name,path,host,port = parts
-                    itemtype = name[0]
-                    name = name[1:]
-                    if port == "70":
-                        port = ""
-                    else:
-                        port = ":%s"%port
-                    if itemtype == "h" and path.startswith("URL:"):
-                        url = path[4:]
-                    else:
-                        url = "gopher://%s%s/%s%s" %(host,port,itemtype,path)
-                    url = url.replace(" ","%20")
-                    linkline = url + " " + name
-                    links.append(linkline)
-                    number = len(links) + startlinks
-                    towrap = "[%s] "%str(number)+ name
-                    r.add_text(towrap)
+                    #If line starts with TAB, there’s no name.
+                    #We thus hide this line
+                    if name:
+                        itemtype = name[0]
+                        name = name[1:]
+                        if port == "70":
+                            port = ""
+                        else:
+                            port = ":%s"%port
+                        if itemtype == "h" and path.startswith("URL:"):
+                            url = path[4:]
+                        else:
+                            url = "gopher://%s%s/%s%s" %(host,port,itemtype,path)
+                        url = url.replace(" ","%20")
+                        linkline = url + " " + name
+                        links.append(linkline)
+                        number = len(links) + startlinks
+                        towrap = "[%s] "%str(number)+ name
+                        r.add_text(towrap)
                 else:
                     r.add_text(line)
         return r.get_final(),links
@@ -870,10 +883,15 @@ class FeedRenderer(GemtextRenderer):
         return "application/rss+xml"
     def is_valid(self):
         if _DO_FEED:
-            parsed = feedparser.parse(self.body)
+            try:
+                parsed = feedparser.parse(self.body)
+            except:
+                parsed = False
         else:
             return False
-        if parsed.bozo:
+        if not parsed:
+            return False
+        elif parsed.bozo:
             return False
         else:
             #If no content, then fallback to HTML
@@ -1312,11 +1330,16 @@ _FORMAT_RENDERERS = {
     "text/gopher": GopherRenderer,
     "image/*": ImageRenderer,
     "application/javascript": HtmlRenderer,
+    "application/json": HtmlRenderer,
+    "text/empty": EmptyRenderer,
 }
 def get_mime(path,url=None):
     #Beware, this one is really a shaddy ad-hoc function
     if not path:
         return None
+    #If the file is empty, simply returns it
+    elif os.path.exists(path) and os.stat(path).st_size == 0:
+        return "text/empty"
     elif url and url.startswith("gopher://"):
         #special case for gopher
         #code copy/pasted from netcache
@@ -1341,7 +1364,7 @@ def get_mime(path,url=None):
         mime = "mailto"
     elif os.path.isdir(path):
         mime = "Local Folder"
-    elif path.endswith(".gmi"):
+    elif path.endswith(".gmi") or path.endswith(".gemini"):
         mime = "text/gemini"
     elif path.endswith("gophermap"):
         mime = "text/gopher"
